from typing import (
    Any,
    Dict,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from cached_property import (
    cached_property,
)
from eth_hash.auto import (
    keccak,
)
from eth_keys.datatypes import (
    PrivateKey,
)
from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    ValidationError,
    to_bytes,
    to_int,
)
import rlp
from rlp.sedes import (
    BigEndianInt,
    Binary,
    CountableList,
    big_endian_int,
    binary,
)

from eth._utils.transactions import (
    calculate_intrinsic_gas,
    create_transaction_signature,
    extract_transaction_sender,
    validate_transaction_signature,
)
from eth.abc import (
    DecodedZeroOrOneLayerRLP,
    ReceiptAPI,
    SignedTransactionAPI,
    TransactionBuilderAPI,
    TransactionDecoderAPI,
    UnsignedTransactionAPI,
)
from eth.constants import (
    CREATE_CONTRACT_ADDRESS,
)
from eth.exceptions import (
    UnrecognizedTransactionType,
)
from eth.rlp.logs import (
    Log,
)
from eth.rlp.receipts import (
    Receipt,
)
from eth.rlp.sedes import (
    address,
)
from eth.rlp.transactions import (
    SignedTransactionMethods,
)
from eth.validation import (
    validate_canonical_address,
    validate_is_bytes,
    validate_is_transaction_access_list,
    validate_uint64,
    validate_uint256,
)
from eth.vm.forks.istanbul.transactions import (
    ISTANBUL_TX_GAS_SCHEDULE,
)
from eth.vm.forks.muir_glacier.transactions import (
    MuirGlacierTransaction,
    MuirGlacierUnsignedTransaction,
)

from .constants import (
    ACCESS_LIST_ADDRESS_COST_EIP_2930,
    ACCESS_LIST_STORAGE_KEY_COST_EIP_2930,
    ACCESS_LIST_TRANSACTION_TYPE,
    VALID_TRANSACTION_TYPES,
)
from .receipts import (
    BerlinReceiptBuilder,
)


class BerlinLegacyTransaction(MuirGlacierTransaction):
    pass


class BerlinUnsignedLegacyTransaction(MuirGlacierUnsignedTransaction):
    def as_signed_transaction(
        self, private_key: PrivateKey, chain_id: int = None
    ) -> BerlinLegacyTransaction:
        v, r, s = create_transaction_signature(self, private_key, chain_id=chain_id)
        return BerlinLegacyTransaction(
            nonce=self.nonce,
            gas_price=self.gas_price,
            gas=self.gas,
            to=self.to,
            value=self.value,
            data=self.data,
            v=v,
            r=r,
            s=s,
        )


class AccountAccesses(rlp.Serializable):
    fields = [
        ("account", address),
        ("storage_keys", CountableList(BigEndianInt(32))),
    ]


class UnsignedAccessListTransaction(rlp.Serializable):
    _type_id = ACCESS_LIST_TRANSACTION_TYPE
    fields = [
        ("chain_id", big_endian_int),
        ("nonce", big_endian_int),
        ("gas_price", big_endian_int),
        ("gas", big_endian_int),
        ("to", address),
        ("value", big_endian_int),
        ("data", binary),
        ("access_list", CountableList(AccountAccesses)),
    ]

    @cached_property
    def _type_byte(self) -> bytes:
        return to_bytes(self._type_id)

    def get_message_for_signing(self) -> bytes:
        payload = rlp.encode(self)
        return self._type_byte + payload

    def as_signed_transaction(self, private_key: PrivateKey) -> "TypedTransaction":
        message = self.get_message_for_signing()
        signature = private_key.sign_msg(message)
        y_parity, r, s = signature.vrs

        signed_transaction = AccessListTransaction(
            self.chain_id,
            self.nonce,
            self.gas_price,
            self.gas,
            self.to,
            self.value,
            self.data,
            self.access_list,
            y_parity,
            r,
            s,
        )
        return TypedTransaction(self._type_id, signed_transaction)

    def validate(self) -> None:
        validate_uint256(self.chain_id, title="Transaction.chain_id")
        validate_uint64(self.nonce, title="Transaction.nonce")
        validate_uint256(self.gas_price, title="Transaction.gas_price")
        validate_uint256(self.gas, title="Transaction.gas")
        if self.to != CREATE_CONTRACT_ADDRESS:
            validate_canonical_address(self.to, title="Transaction.to")
        validate_uint256(self.value, title="Transaction.value")
        validate_is_bytes(self.data, title="Transaction.data")
        validate_is_transaction_access_list(self.access_list)

    def get_intrinsic_gas(self) -> int:
        return _get_access_list_txn_intrinsic_gas(self)

    @property
    def intrinsic_gas(self) -> int:
        return self.get_intrinsic_gas()

    # Old transactions are treated as setting both max-fees as the gas price
    @property
    def max_priority_fee_per_gas(self) -> int:
        return self.gas_price

    @property
    def max_fee_per_gas(self) -> int:
        return self.gas_price


class AccessListTransaction(
    rlp.Serializable, SignedTransactionMethods, SignedTransactionAPI
):
    _type_id = ACCESS_LIST_TRANSACTION_TYPE
    fields = [
        ("chain_id", big_endian_int),
        ("nonce", big_endian_int),
        ("gas_price", big_endian_int),
        ("gas", big_endian_int),
        ("to", address),
        ("value", big_endian_int),
        ("data", binary),
        ("access_list", CountableList(AccountAccesses)),
        ("y_parity", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    ]

    def get_sender(self) -> Address:
        return extract_transaction_sender(self)

    def get_message_for_signing(self) -> bytes:
        unsigned = UnsignedAccessListTransaction(
            self.chain_id,
            self.nonce,
            self.gas_price,
            self.gas,
            self.to,
            self.value,
            self.data,
            self.access_list,
        )
        payload = rlp.encode(unsigned)
        return self._type_byte + payload

    def check_signature_validity(self) -> None:
        validate_transaction_signature(self)

    @cached_property
    def _type_byte(self) -> bytes:
        return to_bytes(self._type_id)

    @cached_property
    def hash(self) -> Hash32:
        raise NotImplementedError("Call hash() on the TypedTransaction instead")

    def get_intrinsic_gas(self) -> int:
        return _get_access_list_txn_intrinsic_gas(self)

    def encode(self) -> bytes:
        return rlp.encode(self)

    def make_receipt(
        self,
        status: bytes,
        gas_used: int,
        log_entries: Tuple[Tuple[bytes, Tuple[int, ...], bytes], ...],
    ) -> ReceiptAPI:
        logs = [Log(address, topics, data) for address, topics, data in log_entries]
        # TypedTransaction is responsible for wrapping this into a TypedReceipt
        return Receipt(
            state_root=status,
            gas_used=gas_used,
            logs=logs,
        )

    # Old transactions are treated as setting both max-fees as the gas price
    @property
    def max_priority_fee_per_gas(self) -> int:
        return self.gas_price

    @property
    def max_fee_per_gas(self) -> int:
        return self.gas_price


class AccessListPayloadDecoder(TransactionDecoderAPI):
    @classmethod
    def decode(cls, payload: bytes) -> SignedTransactionAPI:
        return rlp.decode(payload, sedes=AccessListTransaction)


class TypedTransaction(
    SignedTransactionMethods, SignedTransactionAPI, TransactionDecoderAPI
):
    rlp_type = Binary(min_length=1)  # must have at least one byte for the type
    _inner: SignedTransactionAPI

    decoders: Dict[int, Type[TransactionDecoderAPI]] = {
        ACCESS_LIST_TRANSACTION_TYPE: AccessListPayloadDecoder,
    }
    receipt_builder = BerlinReceiptBuilder

    def __init__(self, type_id: int, proxy_target: SignedTransactionAPI) -> None:
        self.type_id = type_id
        self._inner = proxy_target

    @classmethod
    def get_payload_codec(cls, type_id: int) -> Type[TransactionDecoderAPI]:
        if type_id in cls.decoders:
            return cls.decoders[type_id]
        elif type_id in VALID_TRANSACTION_TYPES:
            raise UnrecognizedTransactionType(type_id, "Unknown transaction type")
        else:
            raise ValidationError(
                f"Cannot build typed transaction with {hex(type_id)} >= 0x80"
            )

    def encode(self) -> bytes:
        return self._type_byte + self._inner.encode()

    @classmethod
    def decode(cls, encoded: bytes) -> SignedTransactionAPI:
        type_id = to_int(encoded[0])
        payload = encoded[1:]

        payload_codec = cls.get_payload_codec(type_id)
        inner_transaction = payload_codec.decode(payload)
        return cls(type_id, inner_transaction)

    @classmethod
    def serialize(cls, obj: "TypedTransaction") -> DecodedZeroOrOneLayerRLP:
        encoded = obj.encode()
        return cls.rlp_type.serialize(encoded)

    @classmethod
    def deserialize(
        cls, encoded_unchecked: DecodedZeroOrOneLayerRLP
    ) -> SignedTransactionAPI:
        # binary checks a few basics, like the length of the bytes
        encoded = cls.rlp_type.deserialize(encoded_unchecked)
        return cls.decode(encoded)

    @cached_property
    def _type_byte(self) -> bytes:
        return to_bytes(self.type_id)

    @property
    def chain_id(self) -> int:
        return self._inner.chain_id

    @property
    def nonce(self) -> int:
        return self._inner.nonce

    @property
    def gas_price(self) -> int:
        return self._inner.gas_price

    @property
    def max_priority_fee_per_gas(self) -> int:
        return self._inner.max_priority_fee_per_gas

    @property
    def max_fee_per_gas(self) -> int:
        return self._inner.max_fee_per_gas

    @property
    def gas(self) -> int:
        return self._inner.gas

    @property
    def to(self) -> Address:
        return self._inner.to

    @property
    def value(self) -> int:
        return self._inner.value

    @property
    def data(self) -> bytes:
        return self._inner.data

    @property
    def y_parity(self) -> int:
        return self._inner.y_parity

    @property
    def r(self) -> int:
        return self._inner.r

    @property
    def s(self) -> int:
        return self._inner.s

    @property
    def access_list(self) -> Sequence[Tuple[Address, Sequence[int]]]:
        return self._inner.access_list

    def get_sender(self) -> Address:
        return self._inner.get_sender()

    def get_message_for_signing(self) -> bytes:
        return self._inner.get_message_for_signing()

    def check_signature_validity(self) -> None:
        self._inner.check_signature_validity()

    @cached_property
    def hash(self) -> Hash32:
        return cast(Hash32, keccak(self.encode()))

    def get_intrinsic_gas(self) -> int:
        return self._inner.get_intrinsic_gas()

    def copy(self, **overrides: Any) -> "TypedTransaction":
        inner_copy = self._inner.copy(**overrides)
        return type(self)(self.type_id, inner_copy)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TypedTransaction):
            return False
        else:
            return self.type_id == other.type_id and self._inner == other._inner

    def make_receipt(
        self,
        status: bytes,
        gas_used: int,
        log_entries: Tuple[Tuple[bytes, Tuple[int, ...], bytes], ...],
    ) -> ReceiptAPI:
        inner_receipt = self._inner.make_receipt(status, gas_used, log_entries)

        return self.receipt_builder.typed_receipt_class(self.type_id, inner_receipt)

    def __hash__(self) -> int:
        return hash((self.type_id, self._inner))


class BerlinTransactionBuilder(TransactionBuilderAPI):
    """
    Responsible for serializing transactions of ambiguous type.

    It dispatches to either the legacy transaction type or the new typed
    transaction, depending on the nature of the encoded/decoded transaction.
    """

    legacy_signed = BerlinLegacyTransaction
    legacy_unsigned = BerlinUnsignedLegacyTransaction
    typed_transaction = TypedTransaction

    @classmethod
    def decode(cls, encoded: bytes) -> SignedTransactionAPI:
        if len(encoded) == 0:
            raise ValidationError(
                "Encoded transaction was empty, which makes it invalid"
            )

        type_id = to_int(encoded[0])
        if type_id in VALID_TRANSACTION_TYPES:
            return cls.typed_transaction.decode(encoded)
        else:
            return rlp.decode(encoded, sedes=cls.legacy_signed)

    @classmethod
    def deserialize(cls, encoded: DecodedZeroOrOneLayerRLP) -> SignedTransactionAPI:
        if isinstance(encoded, bytes):
            return cls.typed_transaction.deserialize(encoded)
        else:
            return cls.legacy_signed.deserialize(encoded)

    @classmethod
    def serialize(cls, obj: SignedTransactionAPI) -> DecodedZeroOrOneLayerRLP:
        if isinstance(obj, cls.typed_transaction):
            return cls.typed_transaction.serialize(obj)
        else:
            return cls.legacy_signed.serialize(obj)

    @classmethod
    def create_unsigned_transaction(
        cls,
        *,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes,
    ) -> UnsignedTransactionAPI:
        return cls.legacy_unsigned(nonce, gas_price, gas, to, value, data)

    @classmethod
    def new_transaction(
        cls,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes,
        v: int,
        r: int,
        s: int,
    ) -> SignedTransactionAPI:
        return cls.legacy_signed(nonce, gas_price, gas, to, value, data, v, r, s)

    @classmethod
    def new_unsigned_access_list_transaction(
        cls,
        chain_id: int,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes,
        access_list: Sequence[Tuple[Address, Sequence[int]]],
    ) -> TypedTransaction:
        transaction = UnsignedAccessListTransaction(
            chain_id,
            nonce,
            gas_price,
            gas,
            to,
            value,
            data,
            access_list,
        )
        return transaction

    @classmethod
    def new_access_list_transaction(
        cls,
        chain_id: int,
        nonce: int,
        gas_price: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes,
        access_list: Sequence[Tuple[Address, Sequence[int]]],
        y_parity: int,
        r: int,
        s: int,
    ) -> TypedTransaction:
        transaction = AccessListTransaction(
            chain_id,
            nonce,
            gas_price,
            gas,
            to,
            value,
            data,
            access_list,
            y_parity,
            r,
            s,
        )
        return cls.typed_transaction(ACCESS_LIST_TRANSACTION_TYPE, transaction)


def _get_access_list_txn_intrinsic_gas(
    klass: Union[UnsignedAccessListTransaction, AccessListTransaction]
) -> int:
    core_gas = calculate_intrinsic_gas(ISTANBUL_TX_GAS_SCHEDULE, klass)

    num_addresses = len(klass.access_list)
    preload_address_costs = ACCESS_LIST_ADDRESS_COST_EIP_2930 * num_addresses

    num_slots = sum(len(slots) for _, slots in klass.access_list)
    preload_slot_costs = ACCESS_LIST_STORAGE_KEY_COST_EIP_2930 * num_slots

    return core_gas + preload_address_costs + preload_slot_costs



import datetime
from typing import (
    Dict,
    Optional,
    Tuple,
)

from eth_typing import (
    Address,
)

from eth.abc import (
    BlockHeaderAPI,
)
from eth.constants import (
    BLANK_ROOT_HASH,
    GAS_LIMIT_ADJUSTMENT_FACTOR,
    GAS_LIMIT_EMA_DENOMINATOR,
    GAS_LIMIT_MAXIMUM,
    GAS_LIMIT_MINIMUM,
    GAS_LIMIT_USAGE_ADJUSTMENT_DENOMINATOR,
    GAS_LIMIT_USAGE_ADJUSTMENT_NUMERATOR,
    GENESIS_BLOCK_NUMBER,
    GENESIS_PARENT_HASH,
    ZERO_ADDRESS,
)
from eth.typing import (
    BlockNumber,
    HeaderParams,
)


def eth_now() -> int:
    """
    The timestamp is in UTC.
    """
    return int(datetime.datetime.utcnow().timestamp())


def new_timestamp_from_parent(parent: Optional[BlockHeaderAPI]) -> int:
    """
    Generate a timestamp to use on a new header.

    Generally, attempt to use the current time. If timestamp is too old (equal
    or less than parent), return `parent.timestamp + 1`. If parent is None,
    then consider this a genesis block.
    """
    if parent is None:
        return eth_now()
    else:
        # header timestamps must increment
        return max(
            parent.timestamp + 1,
            eth_now(),
        )


def fill_header_params_from_parent(
    parent: BlockHeaderAPI,
    gas_limit: int,
    difficulty: int,
    timestamp: int,
    coinbase: Address = ZERO_ADDRESS,
    nonce: bytes = None,
    extra_data: bytes = None,
    transaction_root: bytes = None,
    state_root: bytes = None,
    mix_hash: bytes = None,
    receipt_root: bytes = None,
) -> Dict[str, HeaderParams]:
    if parent is None:
        parent_hash = GENESIS_PARENT_HASH
        block_number = GENESIS_BLOCK_NUMBER
        if state_root is None:
            state_root = BLANK_ROOT_HASH
    else:
        parent_hash = parent.hash
        block_number = BlockNumber(parent.block_number + 1)

        if state_root is None:
            state_root = parent.state_root

    header_kwargs: Dict[str, HeaderParams] = {
        "parent_hash": parent_hash,
        "coinbase": coinbase,
        "state_root": state_root,
        "gas_limit": gas_limit,
        "difficulty": difficulty,
        "block_number": block_number,
        "timestamp": timestamp,
    }
    if nonce is not None:
        header_kwargs["nonce"] = nonce
    if extra_data is not None:
        header_kwargs["extra_data"] = extra_data
    if transaction_root is not None:
        header_kwargs["transaction_root"] = transaction_root
    if receipt_root is not None:
        header_kwargs["receipt_root"] = receipt_root
    if mix_hash is not None:
        header_kwargs["mix_hash"] = mix_hash

    return header_kwargs


def compute_gas_limit_bounds(previous_limit: int) -> Tuple[int, int]:
    """
    Compute the boundaries for the block gas limit based on the parent block.
    """
    boundary_range = previous_limit // GAS_LIMIT_ADJUSTMENT_FACTOR

    # the boundary range is the exclusive limit, therefore the inclusive bounds are
    # (boundary_range - 1) and (boundary_range + 1) for upper and lower bounds, respectively  # noqa: E501
    upper_bound_inclusive = min(GAS_LIMIT_MAXIMUM, previous_limit + boundary_range - 1)
    lower_bound_inclusive = max(GAS_LIMIT_MINIMUM, previous_limit - boundary_range + 1)
    return lower_bound_inclusive, upper_bound_inclusive


def compute_gas_limit(parent_header: BlockHeaderAPI, genesis_gas_limit: int) -> int:
    """
    A simple strategy for adjusting the gas limit.

    For each block:

    - decrease by 1/1024th of the gas limit from the previous block
    - increase by 50% of the total gas used by the previous block

    If the value is less than the given `genesis_gas_limit`:

    - increase the gas limit by 1/1024th of the gas limit from the previous block.

    If the value is less than the GAS_LIMIT_MINIMUM:

    - use the GAS_LIMIT_MINIMUM as the new gas limit.
    """
    if genesis_gas_limit < GAS_LIMIT_MINIMUM:
        raise ValueError(
            "The `genesis_gas_limit` value must be greater than the "
            f"GAS_LIMIT_MINIMUM.  Got {genesis_gas_limit}.  Must be greater than "
            f"{GAS_LIMIT_MINIMUM}"
        )

    if parent_header is None:
        return genesis_gas_limit

    decay = parent_header.gas_limit // GAS_LIMIT_EMA_DENOMINATOR

    if parent_header.gas_used:
        usage_increase = (
            (parent_header.gas_used * GAS_LIMIT_USAGE_ADJUSTMENT_NUMERATOR)
            // (GAS_LIMIT_USAGE_ADJUSTMENT_DENOMINATOR)
            // (GAS_LIMIT_EMA_DENOMINATOR)
        )
    else:
        usage_increase = 0

    gas_limit = max(
        GAS_LIMIT_MINIMUM,
        # + 1 because the decay is an exclusive limit we have to remain inside of
        (parent_header.gas_limit - decay + 1) + usage_increase,
    )

    if gas_limit < GAS_LIMIT_MINIMUM:
        return GAS_LIMIT_MINIMUM
    elif gas_limit < genesis_gas_limit:
        # - 1 because the decay is an exclusive limit we have to remain inside of
        return parent_header.gas_limit + decay - 1
    else:
        return gas_limit



from eth_utils.toolz import (
    merge,
)

from eth import (
    precompiles,
)
from eth._utils.address import (
    force_bytes_to_address,
)
from eth.vm.forks.petersburg.computation import (
    PETERSBURG_PRECOMPILES,
    PetersburgComputation,
)
from eth.vm.gas_meter import (
    GasMeter,
    allow_negative_refund_strategy,
)

from .constants import (
    GAS_ECADD,
    GAS_ECMUL,
    GAS_ECPAIRING_BASE,
    GAS_ECPAIRING_PER_POINT,
)
from .opcodes import (
    ISTANBUL_OPCODES,
)

ISTANBUL_PRECOMPILES = merge(
    PETERSBURG_PRECOMPILES,
    {
        force_bytes_to_address(b"\x06"): precompiles.ecadd(gas_cost=GAS_ECADD),
        force_bytes_to_address(b"\x07"): precompiles.ecmul(gas_cost=GAS_ECMUL),
        force_bytes_to_address(b"\x08"): precompiles.ecpairing(
            gas_cost_base=GAS_ECPAIRING_BASE,
            gas_cost_per_point=GAS_ECPAIRING_PER_POINT,
        ),
        force_bytes_to_address(b"\x09"): precompiles.blake2b_fcompress,
    },
)


class IstanbulComputation(PetersburgComputation):
    """
    A class for all execution *message* computations in the ``Istanbul`` fork.
    Inherits from
    :class:`~eth.vm.forks.constantinople.petersburg.PetersburgComputation`
    """

    # Override
    opcodes = ISTANBUL_OPCODES
    _precompiles = ISTANBUL_PRECOMPILES

    def _configure_gas_meter(self) -> GasMeter:
        return GasMeter(self.msg.gas, allow_negative_refund_strategy)



from collections import (
    defaultdict,
)
from typing import (
    Any,
    Dict,
    List,
)

from eth_utils import (
    encode_hex,
)

from eth.tools._utils.mappings import (
    deep_merge,
)
from eth.tools._utils.normalization import (
    normalize_environment,
    normalize_networks,
    normalize_state,
    normalize_transaction_group,
)
from eth.tools.fixtures.helpers import (
    get_test_name,
)
from eth.vm.forks.byzantium.state import (
    ByzantiumState,
)
from eth.vm.forks.constantinople.state import (
    ConstantinopleState,
)
from eth.vm.forks.frontier.state import (
    FrontierState,
)
from eth.vm.forks.homestead.state import (
    HomesteadState,
)
from eth.vm.forks.istanbul.state import (
    IstanbulState,
)
from eth.vm.forks.petersburg.state import (
    PetersburgState,
)
from eth.vm.forks.spurious_dragon.state import (
    SpuriousDragonState,
)
from eth.vm.forks.tangerine_whistle.state import (
    TangerineWhistleState,
)

from ._utils import (
    calc_state_root,
)

ALL_NETWORKS = [
    "Frontier",
    "Homestead",
    "EIP150",
    "EIP158",
    "Byzantium",
    "Constantinople",
    "Petersburg",
    "Istanbul",
]

STATE_CLASSES = {
    "Frontier": FrontierState,
    "Homestead": HomesteadState,
    "EIP150": TangerineWhistleState,
    "EIP158": SpuriousDragonState,
    "Byzantium": ByzantiumState,
    "Constantinople": ConstantinopleState,
    "Petersburg": PetersburgState,
    "Istanbul": IstanbulState,
}

_missing_state_classes = set(ALL_NETWORKS) - set(STATE_CLASSES)
assert not any(_missing_state_classes)


def fill_state_test(filler: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
    """
    Filler function for filling state tests.
    """
    test_name = get_test_name(filler)
    test = filler[test_name]

    environment = normalize_environment(test["env"])
    pre_state = normalize_state(test["pre"])
    transaction_group = normalize_transaction_group(test["transaction"])

    post: Dict[int, List[Dict[str, str]]] = defaultdict(list)
    for expect in test["expect"]:
        indexes = expect["indexes"]
        networks = normalize_networks(expect["networks"])
        result = normalize_state(expect["result"])
        post_state = deep_merge(pre_state, result)
        for network in networks:
            state_class = STATE_CLASSES[network]
            post_state_root = calc_state_root(post_state, state_class)
            post[network].append(
                {
                    "hash": encode_hex(post_state_root),
                    "indexes": indexes,
                }
            )

    return {
        test_name: {
            "env": environment,
            "pre": pre_state,
            "transaction": transaction_group,
            "post": post,
        }
    }



from typing import (
    Iterable,
    Optional,
)

from eth_typing import (
    Address,
    BlockNumber,
    Hash32,
)

from eth._utils.generator import (
    CachedIterable,
)
from eth.abc import (
    ExecutionContextAPI,
)


class ExecutionContext(ExecutionContextAPI):
    _coinbase = None
    _timestamp = None
    _number = None
    _difficulty = None
    _mix_hash = None
    _gas_limit = None
    _prev_hashes = None
    _chain_id = None
    _base_fee_per_gas = None

    def __init__(
        self,
        coinbase: Address,
        timestamp: int,
        block_number: BlockNumber,
        difficulty: int,
        mix_hash: Hash32,
        gas_limit: int,
        prev_hashes: Iterable[Hash32],
        chain_id: int,
        base_fee_per_gas: Optional[int] = None,
    ) -> None:
        self._coinbase = coinbase
        self._timestamp = timestamp
        self._block_number = block_number
        self._difficulty = difficulty
        self._mix_hash = mix_hash
        self._gas_limit = gas_limit
        self._prev_hashes = CachedIterable(prev_hashes)
        self._chain_id = chain_id
        self._base_fee_per_gas = base_fee_per_gas

    @property
    def coinbase(self) -> Address:
        return self._coinbase

    @property
    def timestamp(self) -> int:
        return self._timestamp

    @property
    def block_number(self) -> BlockNumber:
        return self._block_number

    @property
    def difficulty(self) -> int:
        return self._difficulty

    @property
    def mix_hash(self) -> Hash32:
        return self._mix_hash

    @property
    def gas_limit(self) -> int:
        return self._gas_limit

    @property
    def prev_hashes(self) -> Iterable[Hash32]:
        return self._prev_hashes

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def base_fee_per_gas(self) -> int:
        if self._base_fee_per_gas is None:
            raise AttributeError(
                f"This header at Block #{self.block_number} "
                "does not have a base gas fee"
            )
        else:
            return self._base_fee_per_gas



from eth_typing import (
    Hash32,
)


class PyEVMError(Exception):
    """
    Base class for all py-evm errors.
    """

    pass


class VMNotFound(PyEVMError):
    """
    Raised when no VM is available for the provided block number.
    """

    pass


class StateRootNotFound(PyEVMError):
    """
    Raised when the requested state root is not present in our DB.
    """

    @property
    def missing_state_root(self) -> Hash32:
        return self.args[0]


class HeaderNotFound(PyEVMError):
    """
    Raised when a header with the given number/hash does not exist.
    """


class BlockNotFound(PyEVMError):
    """
    Raised when the block with the given number/hash does not exist.
    This will happen, for example, if the transactions or uncles are not
    saved in the database.
    """

    pass


class TransactionNotFound(PyEVMError):
    """
    Raised when the transaction with the given hash or block index does not exist.
    """

    pass


class UnrecognizedTransactionType(PyEVMError):
    """
    Raised when an encoded transaction is using a first byte that is valid, but
    unrecognized. According to EIP 2718, the byte may be in the range [0, 0x7f].
    As of the Berlin hard fork, all of those versions are undefined, except for
    0x01 in EIP 2930.
    """

    @property
    def type_int(self) -> int:
        return self.args[0]


class ReceiptNotFound(PyEVMError):
    """
    Raised when the Receipt with the given receipt index does not exist.
    """

    pass


class ParentNotFound(HeaderNotFound):
    """
    Raised when the parent of a given block does not exist.
    """

    pass


class CanonicalHeadNotFound(PyEVMError):
    """
    Raised when the chain has no canonical head.
    """

    pass


class GapTrackingCorrupted(PyEVMError):
    """
    Raised when the tracking of chain gaps appears to be corrupted
    (e.g. overlapping gaps)
    """

    pass


class CheckpointsMustBeCanonical(PyEVMError):
    """
    Raised when a persisted header attempts to de-canonicalize a checkpoint
    """

    pass


class Halt(PyEVMError):
    """
    Raised when an opcode function halts vm execution.
    """

    pass


class VMError(PyEVMError):
    """
    Base class for errors raised during VM execution.
    """

    burns_gas = True
    erases_return_data = True


class OutOfGas(VMError):
    """
    Raised when a VM execution has run out of gas.
    """

    pass


class InsufficientStack(VMError):
    """
    Raised when the stack is empty.
    """

    pass


class FullStack(VMError):
    """
    Raised when the stack is full.
    """

    pass


class InvalidJumpDestination(VMError):
    """
    Raised when the jump destination for a JUMPDEST operation is invalid.
    """

    pass


class InvalidInstruction(VMError):
    """
    Raised when an opcode is invalid.
    """

    pass


class InsufficientFunds(VMError):
    """
    Raised when an account has insufficient funds to transfer the
    requested value.
    """

    pass


class StackDepthLimit(VMError):
    """
    Raised when the call stack has exceeded it's maximum allowed depth.
    """

    pass


class ContractCreationCollision(VMError):
    """
    Raised when there was an address collision during contract creation.
    """

    pass


class IncorrectContractCreationAddress(VMError):
    """
    Raised when the address provided by transaction does not
    match the calculated contract creation address.
    """

    pass


class Revert(VMError):
    """
    Raised when the REVERT opcode occured
    """

    burns_gas = False
    erases_return_data = False


class WriteProtection(VMError):
    """
    Raised when an attempt to modify the state database is made while
    operating inside of a STATICCALL context.
    """

    pass


class OutOfBoundsRead(VMError):
    """
    Raised when an attempt was made to read data beyond the
    boundaries of the buffer (such as with RETURNDATACOPY)
    """

    pass


class ReservedBytesInCode(VMError):
    """
    Raised when bytes for the code to be deployed are reserved
    for a particular reason.
    """

    pass
